import logging
import time

from prompt_toolkit.application import get_app
from prompt_toolkit.formatted_text import HTML, to_formatted_text, fragment_list_to_text
from prompt_toolkit.layout.processors import Processor, Transformation
from prompt_toolkit.layout import BufferControl, ConditionalContainer
from prompt_toolkit.buffer import Buffer
from prompt_toolkit.layout.layout import Layout
from prompt_toolkit.styles import Style

from prompt_toolkit.filters import Condition
from prompt_toolkit.layout.controls import FormattedTextControl

from prompt_toolkit.layout.containers import HSplit, VSplit, Window, WindowAlign
from prompt_toolkit.widgets import TextArea

from prompt_toolkit.auto_suggest import AutoSuggestFromHistory


_log = logging.getLogger("katcpcli.UI")

# Style.
tui_style = Style(
    [
        ("output-field", "bg:#000044 #ffffff"),
        ("status", "bg:#000000 #ffffff"),
        ("line", "#004400"),
        ("connected", "fg:#00cf00"),
        ("disconnected", "fg:#cf0000"),
    ]
)

prompt_style = Style(
    [
        ("output-field", "bg:#000044 #ffffff"),
        ("status", "bg:#000000 #ffffff"),
        ("line", "#004400"),
        ("connected", "bg:#00cf00"),
        ("disconnected", "bg:#cf0000"),
    ]
)


def print_help(dispatcher, kb):
    """
    write help to the dispatcher
    """
    dispatcher.output_stream("")
    dispatcher.output_stream(
        "The following commands and key bindings are supported by the application. [?] can be used as a shortcut for katpcmd."
    )
    dispatcher.output_stream("")
    dispatcher.output_stream(HTML("<b>Commands:</b>"))
    for c, v in dispatcher.commands.items():
        desc = v.__doc__.lstrip().rstrip()
        dispatcher.output_stream(f"{c:15} - {desc}")

    dispatcher.output_stream("")
    dispatcher.output_stream(HTML("<b>Key Bindings:</b>"))
    for b in kb.bindings:
        keys = ",".join(k.value for k in b.keys)
        desc = b.handler.__doc__

        keys = keys.replace("escape,", "alt+")

        dispatcher.output_stream(f"{keys:15} - {desc}")


def _bouble_gen():
    while True:
        for s in [".", "o", "O"]:
            yield s


bouble = _bouble_gen()


def bottom_toolbar(dispatcher):
    """
    Render right bottom toolbar.
    """
    style = "disconnected"
    host = "-"
    port = "-"
    if dispatcher.katcpclient:
        host = dispatcher.katcpclient.host
        port = dispatcher.katcpclient.port
        if dispatcher.katcpclient.is_connected:
            style = "connected"

    # Indicate that a task is being processed
    if dispatcher.running_tasks:
        active = next(bouble)
    else:
        active = " "

    return HTML(
        f"<b>KATCP cli</b> connected to (<{style}><b>{host}:{port}</b></{style}>) [{active}] -  Use [Ctrl-C] to terminate, [F1] for help"
    )


def rprompt():
    """
    Render right prompt.
    """
    return HTML(f"<style color='green'>{time.ctime()}</style>")


class Buffer_(Buffer):
    """
    Buffer that handles formatted text. See prompt_toolkit issue #711
    https://github.com/prompt-toolkit/python-prompt-toolkit/issues/711
    """

    def write(self, data):
        """
        Write data as a new line to the buffer. Mimics the stream interface to enable usage in log handler.
        """
        if isinstance(data, HTML):
            line = data.value
        else:
            line = str(data)

        self.text += line.rstrip() + "\n"

        if not get_app().layout.has_focus(self):
            self.cursor_position = len(self.text)

    def flush(self):
        """
        No op as flushing is handled by toolkit.
        """


class FormatText(Processor):  # pylint: disable=too-few-public-methods
    """
    Formatter for buffer to handle formatted text. See prompt_toolkit issue #711
    https://github.com/prompt-toolkit/python-prompt-toolkit/issues/711
    """

    def apply_transformation(self, transformation_input):
        try:
            fragments = to_formatted_text(
                HTML(fragment_list_to_text(transformation_input.fragments))
            )
        except Exception as E:
            return Transformation(transformation_input.fragments)
        return Transformation(fragments)


class FSApplicationUI:
    """
    User-interface for full screen application
    """

    def __init__(self, completer, history):

        self.show_log_window = False

        self.bottom_toolbar_generator = None

        # The Buffers for the log and output stream
        self.log_field = Buffer_()
        self.output_field = Buffer_()

        self.input_field = TextArea(
            height=1,
            completer=completer,
            auto_suggest=AutoSuggestFromHistory(),
            history=history,
            prompt=">>> ",
            style="class:input-field",
            multiline=False,
            wrap_lines=False,
        )

        # Building the application layout
        container = HSplit(
            [
                Window(
                    FormattedTextControl("Log Informs (Toggle using F2)"),
                    style="class:status",
                    char="-",
                    height=1,
                ),
                ConditionalContainer(
                    content=Window(
                        BufferControl(
                            self.log_field,
                            input_processors=[FormatText()],
                            include_default_input_processors=True,
                        ),
                        height=6,
                        wrap_lines=True,
                    ),
                    filter=Condition(lambda: self.show_log_window),
                ),
                Window(
                    FormattedTextControl("Output"),
                    style="class:status",
                    height=1,
                    char="-",
                ),
                Window(
                    BufferControl(
                        self.output_field,
                        input_processors=[FormatText()],
                        include_default_input_processors=True,
                    ),
                    wrap_lines=True,
                ),
                VSplit(
                    [
                        Window(
                            FormattedTextControl(self.bottom_toolbar),
                            style="class:status",
                            height=1,
                        ),
                        Window(
                            FormattedTextControl(rprompt),
                            style="class:status",
                            height=1,
                            align=WindowAlign.RIGHT,
                            dont_extend_width=True,
                        ),
                    ]
                ),
                self.input_field,
            ]
        )

        self.layout = Layout(container, focused_element=self.input_field)

        # TUI logs to log window
        logging.root.handlers.clear()
        sh = logging.StreamHandler(stream=self.log_field)
        sh.setFormatter(
            logging.Formatter(
                "%(asctime)s - %(name)s - %(filename)s:"
                "%(lineno)s - %(levelname)s - %(message)s"
            )
        )
        logging.getLogger('').addHandler(sh)

    def bottom_toolbar(self):
        """
        Return bottom toolbar content
        """
        if self.bottom_toolbar_generator is None:
            return ""
        return self.bottom_toolbar_generator()




import unittest
import unittest.mock
import sys
import socket
import asyncio
import time
import tempfile
import logging
import os

from prompt_toolkit.shortcuts import PromptSession
from prompt_toolkit.input import create_pipe_input
from prompt_toolkit.application import create_app_session
from prompt_toolkit.shortcuts import print_formatted_text
from prompt_toolkit.output import DummyOutput

import aiokatcp
from katcpcli.StreamClient import sensor_status_format
from katcpcli.CommandDispatcher import CommandDispatcher
from katcpcli import app, config, main

sys.path.insert(0, '..')


class MockServer(aiokatcp.DeviceServer):
    VERSION = "mock-1.0"
    BUILD_STATE = 'mock-1.0'

    def __init__(self):
        with socket.socket() as sock:
            sock.bind(('', 0))
            self.port = sock.getsockname()[1]

        aiokatcp.DeviceServer.__init__(self, '127.0.0.1', self.port)
        self.mock = unittest.mock.Mock()

    async def request_echo(self, ctx, line):
        """
        foo doc
        """
        self.mock.echo(line)
        return line

    async def request_sendloginform(self, ctx, msg):
        """
        Send a log-inform
        """
        self.mass_inform("log", "info", time.time(), f"log inform at info level with {msg.decode('ascii')}")




class SimpleBuffer:
    def __init__(self):
        self.data = []

    def __call__(self, line):
        self.data.append(line)



class TestFormattingHelpers(unittest.TestCase):
    def test_sensor_status_format(self):
        """
        a valid string should be returned for all posible status values.
        """
        for v in aiokatcp.Sensor.Status:
            self.assertTrue(isinstance(sensor_status_format(v), str), msg=f'Error with {v}')

        for v in aiokatcp.Sensor.Status:
            self.assertTrue(isinstance(sensor_status_format(v.value), str), msg=f'Error with {v}')

        for v in aiokatcp.Sensor.Status:
            self.assertTrue(isinstance(sensor_status_format(v.name), str), msg=f'Error with {v}')


class TestCommandDispatcher(unittest.IsolatedAsyncioTestCase):
    def setUp(self):

        self.ostream = SimpleBuffer()
        self.lstream = SimpleBuffer()

        self.server = MockServer()
        self.dispatcher = CommandDispatcher(self.ostream, self.lstream)

    async def asyncTearDown(self):
        await self.server.stop()

    async def asyncSetUp(self):
        await self.server.start()
        await self.dispatcher.connect('127.0.0.1', self.server.port)

    async def testCompletionAfterConnect(self):
        """
        After connect the compeltion set should contain the requests of the server.
        """
        self.assertIn('?echo', self.dispatcher.completion_set)

    async def testRequestDispatch(self):
        await self.dispatcher("?echo foo")
        await asyncio.sleep(.1)
        self.server.mock.echo.assert_called_once()
        # The echo line should be the end of the last line in the ostream
        self.assertTrue(self.ostream.data[-1].value.endswith('foo'))

    async def testUnescapedHTMLCommand(self):
        # special characters should not raise an exception but onlyc reate an
        # error
        await self.dispatcher("<")
        await self.dispatcher(">")

    async def testSingleQuotationMarks(self):
        # single quotation amrks should not raise an exception but only an
        # error message
        await self.dispatcher("'")
        await self.dispatcher('"')


    async def testLogInformToggle(self):
        # Log informs should be supressed
        await self.dispatcher.log_inform('enabled')
        self.assertEqual(self.dispatcher.supressed_logs, 0)
        await self.dispatcher("?sendloginform foo")
        await asyncio.sleep(.1)
        self.assertTrue(self.lstream.data[-1].value.endswith('foo'))
        self.assertIn("log inform at info level", self.lstream.data[-1].value)

        await self.dispatcher.log_inform('disabled')
        await self.dispatcher("?sendloginform bar")
        await asyncio.sleep(.1)
        self.assertFalse(self.lstream.data[-1].value.endswith('bar'))
        self.assertEqual(self.dispatcher.supressed_logs, 1)


    async def testConnect(self):
        "Connect should work with host:port and host port"
        await self.dispatcher.connect('127.0.0.1', self.server.port)
        self.assertTrue(self.ostream.data[-1].endswith('established.'))

        self.ostream.data.clear()
        await self.dispatcher.connect(f'127.0.0.1:{self.server.port}')
        self.assertTrue(self.ostream.data[-1].endswith('established.'))

        # invalid connect string should give error message and no exception
        await self.dispatcher.connect(f'127.0.0.1')
        self.assertIn('Error', self.ostream.data[-1].value)


class TestConfig(unittest.TestCase):
    def test_non_existin_config_file(self):
        # should just ignore this and return default config
        cfg = config.from_file('/path/to/non/existing_file_foo_bar')
        self.assertEqual(len(cfg['favorites']), 0)

    def test_existing_file_favourites(self):
        with tempfile.NamedTemporaryFile(mode='w') as f:
            f.write("""
            [favorites]
            foo=bar
            spam=eggs
            """)
            f.flush()
            cfg = config.from_file(f.name)

            self.assertEqual(cfg['favorites']['foo'], 'bar')
            self.assertEqual(cfg['favorites']['spam'], 'eggs')

    def testDevice2HostPort(self):

        favorites = {'foo': 'bar:123'}

        host, port = main.device2hostport('spam:1234', favorites)
        self.assertEqual(host, 'spam')
        self.assertEqual(port, '1234')

        host, port = main.device2hostport('ham', favorites)
        self.assertEqual(host, 'ham')

        host, port = main.device2hostport('foo', favorites)
        self.assertEqual(host, 'bar')
        self.assertEqual(port, '123')


if __name__ == "__main__":
    import coloredlogs

    logging.getLogger().addHandler(logging.NullHandler())
    log = logging.getLogger('katcpcli')
    if os.environ.get('VERBOSE', '').upper() == 'TRUE':
        log.setLevel(logging.DEBUG)
        level = logging.DEBUG
    else:
        log.setLevel(logging.INFO)
        level = logging.CRITICAL

    coloredlogs.install(
        fmt=("[ %(levelname)s - %(asctime)s - %(name)s "
             "- %(filename)s:%(lineno)s] %(message)s"),
        level=level,            # We manage the log level via the logger, not the handler
        logger=log)

    unittest.main()





#!/usr/bin/python3
"""
KATCPcli - a text client for KATCP devices.
"""

import asyncio
import logging
import sys
import argparse
import os

import katcpcli.app
import katcpcli.config

logging.basicConfig(
    level=logging.INFO,
    stream=sys.stderr,
    format="%(asctime)s - %(name)s - %(filename)s:"
    "%(lineno)s - %(levelname)s - %(message)s",
)

_log = logging.getLogger("katcpcli")

DEFAULT_HOST = "localhost"
DEFAULT_PORT = 1235


def device2hostport(device, favorites):
    """
    Split a device string into host, port. considering the default port and the favorites
    """
    if ":" in device:
        _log.debug('Port and host split')
        host, port = device.split(":")
    elif device in favorites:
        _log.debug('Device found in favorites')
        host, port = favorites[device].split(':')
    else:
        _log.debug('Using devce only')
        host = device
        port = DEFAULT_PORT

    return host, port


async def async_main():
    """
    Run katcpcli program.
    """
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument(
            "--cfgfile",
            default=os.path.join(os.path.expanduser('~'), '.katcpclirc'),
            help="Config file to use. Default: ~/.katcpclirc",
        )
    parser.add_argument(
        "-a",
        "--host",
        help="Host Device to connect to. Updates device setting if both are given.",
    )
    parser.add_argument(
        "-p",
        "--port",
        help="Port to connect to. Updates device setting if both are given.",
    )
    parser.add_argument(
        "-f",
        "--favorites",
        help="Show favorites from config",
        action='store_true'
    )
    parser.add_argument(
        "device",
        nargs="?",
        help="Device to connect to [ip:port]",
        default=f"{DEFAULT_HOST}:{DEFAULT_PORT}",
    )

    parser.add_argument(
        "-v", "--verbose", action="store_true", help="Print debug output"
    )
    parser.add_argument("--nofs", action="store_true", help="Start as simple prompt app.")

    args = parser.parse_args()
    if args.verbose:
        _log.setLevel("DEBUG")

    cfg = katcpcli.config.from_file(os.path.abspath(args.cfgfile))

    if args.favorites:
        print('Configured favorites:')
        for k, v in cfg['favorites'].items():
            print(f'  * {k:30} --> {v}')
        sys.exit(0)

    host, port = device2hostport(args.device, cfg['favorites'])

    if args.host:
        host = args.host
    if args.port:
        port = args.port

    dispatcher, app = katcpcli.app.create_app_and_dispatcher(args.nofs)
    asyncio.create_task(dispatcher.connect(host, port))

    print("Starting katcp CLI ... ")
    await app.run_async()


def main():
    """
    Run asyncio main
    """
    asyncio.run(async_main())
    print("Exiting katcp CLI ...")


if __name__ == "__main__":
    main()



import configparser
import os
import logging

_log = logging.getLogger("katcpcli")


def default_config():
    config = configparser.ConfigParser()
    config.add_section('favorites')
    return config


def from_file(path):
    _log.debug('Reading config from file: %s', path)

    config = default_config()
    if not os.path.isfile(path):
        _log.debug("No such file: %s", path)
        return config

    try:
        config.read([path])
    except configparser.Error as E:
        _log.error("Cannot parse config file: %s", E)

    return config




import asyncio
import time
import logging
import sys
import os

from prompt_toolkit import PromptSession
from prompt_toolkit.auto_suggest import AutoSuggestFromHistory
from prompt_toolkit.completion import WordCompleter
from prompt_toolkit.history import FileHistory
from prompt_toolkit import print_formatted_text
from prompt_toolkit.keys import Keys 
from prompt_toolkit.key_binding.bindings.page_navigation import (
    scroll_page_up,
    scroll_page_down,
    scroll_forward,
    scroll_backward
)

from prompt_toolkit.application import Application, get_app
from prompt_toolkit.key_binding import KeyBindings


from .CommandDispatcher import CommandDispatcher
from . import UI


_log = logging.getLogger("katcpcli.app")


HISTORY_FILE = os.path.join(os.path.expanduser("~"), ".katcpcli2_history")
REFRESH_RATE = 0.5


def create_app_and_dispatcher(prompt_application):
    """
    Create dispatcher and app, either full screen TUI or prompt app.
    """
    dispatcher = CommandDispatcher(print_formatted_text, print_formatted_text)

    kb = KeyBindings()

    @kb.add("c-c")
    @kb.add("c-d")
    @kb.add("c-q")
    async def _quit(*args):  # pylint: disable=unused-argument
        "Exit the application"
        _log.debug("Exit called")
        get_app().exit()

    @kb.add("f1")
    async def _help(*args):  # pylint: disable=unused-argument
        "Print help"
        UI.print_help(dispatcher, kb)

    dispatcher.commands["help"] = _help
    dispatcher.commands["quit"] = _quit
    dispatcher.commands["exit"] = _quit

    if prompt_application:
        app = prompt_app(dispatcher, kb)
    else:
        app = fs_app(dispatcher, kb)

    return dispatcher, app


def input_handler(buff, dispatcher):
    """
    Echos the input to the output stream and forward it to the dispatcher
    """
    if not buff.document.text:
        _log.debug("Received empty input")
        return
    ts = time.ctime()
    width = (get_app().output.get_size().columns or 80) - len(ts) - 4
    _log.debug("Received input, buffer width: %i", width)
    line_template = f">>> {{out:{width}}}{{ts}}\n"
    dispatcher.output_stream(line_template.format(out=buff.document.text, ts=ts))
    asyncio.create_task(dispatcher(buff.document.text))


def prompt_app(
    dispatcher, kb, history_file=HISTORY_FILE, refresh_interval=REFRESH_RATE
):
    """
    Create and return prompt application
    """

    completer = WordCompleter(words=[], WORD=True)
    completer.words = dispatcher.completion_set

    session = PromptSession(
        ">>> ",
        history=FileHistory(history_file),
        auto_suggest=AutoSuggestFromHistory(),
        completer=completer,
        complete_in_thread=True,
        key_bindings=kb,
        rprompt=UI.rprompt,
        bottom_toolbar=lambda: UI.bottom_toolbar(dispatcher),
        refresh_interval=refresh_interval,
        style=UI.prompt_style,
        multiline=False,
    )

    session.default_buffer.accept_handler = lambda buff: input_handler(buff, dispatcher)

    return session.app


def fs_app(dispatcher, kb, history_file=HISTORY_FILE, refresh_interval=REFRESH_RATE):
    """
    Create and return full screen TUI application
    """
    completer = WordCompleter(words=[], WORD=True)
    completer.words = dispatcher.completion_set
    tui = UI.FSApplicationUI(completer, FileHistory(history_file))
    tui.bottom_toolbar_generator = lambda: UI.bottom_toolbar(dispatcher)

    dispatcher.output_stream = tui.output_field.write
    dispatcher.log_stream = tui.log_field.write

    tui.input_field.accept_handler = lambda buff: input_handler(buff, dispatcher)

    @kb.add("f2")
    def _(_):  # pylint: disable=unused-argument
        "Toggle log window visibility"
        tui.show_log_window = not tui.show_log_window

    @kb.add("pageup")
    def _(event):
        "Scroll up output window"
        w = event.app.layout.current_window
        event.app.layout.focus(tui.output_field)
        scroll_page_up(event)
        event.app.layout.focus(w)

    @kb.add("pagedown")
    def _(event):
        "Scroll up output window"
        w = event.app.layout.current_window
        event.app.layout.focus(tui.output_field)
        scroll_page_down(event)
        event.app.layout.focus(w)

    @kb.add("escape", "pageup")
    def _(event):
        "Scroll down log window"
        w = event.app.layout.current_window
        event.app.layout.focus(tui.log_field)
        scroll_page_up(event)
        event.app.layout.focus(w)

    @kb.add("escape", "pagedown")
    def _(event):
        "Scroll down log window"
        w = event.app.layout.current_window
        event.app.layout.focus(tui.log_field)
        scroll_page_down(event)
        event.app.layout.focus(w)

    @kb.add(Keys.Vt100MouseEvent)
    def _(event):
        w = event.app.layout.current_window
        event.app.layout.focus(tui.output_field)
        if event.data[4] == '4':
            scroll_backward(event)
        elif event.data[4] == '5':
            scroll_forward(event)

        event.app.layout.focus(w)

    application = Application(
        layout=tui.layout,
        key_bindings=kb,
        mouse_support=False,
        full_screen=True,
        style=UI.tui_style,
        refresh_interval=refresh_interval,
    )

    return application
